

# DEVICE 1.2.4


# Imports
import collections
import datetime
import glob
import ipaddress
import os
import platform
import socket
from ctypes import *
from . import Error


# Class Device
class Device:

    # DHCP Status
    DHCP_DISABLED = 0
    DHCP_ENABLED = 1

    # Get Library List
    _lib = glob.glob(os.path.join(os.path.dirname(__file__), 'lib', '*.dll' if (platform.system() == 'Windows')
        else ('*.dylib' if (platform.system() == 'Darwin') else '*.so')))

    # Iterate Library List
    for lib in _lib:

        try:

            # Load Library
            _lib = cdll.LoadLibrary(lib)
            break

        except OSError:

            # Ignore Error
            pass

    # Check Library
    if isinstance(_lib, list):

        # Error
        raise Error.Failed

    # Configure Library
    _lib.burn.argtypes = [c_uint, c_char_p, c_uint]
    _lib.burn.restype = c_ubyte
    _lib.configure.argtypes = [c_uint, c_char_p, c_uint, c_ubyte, c_uint, c_uint, c_uint]
    _lib.configure.restype = c_ubyte
    _lib.find.argtypes = [POINTER(c_uint), POINTER(c_uint), POINTER(c_uint), c_uint]
    _lib.find.restype = c_ubyte
    _lib.get_driver.argtypes = [c_char_p, POINTER(c_uint), c_char_p, POINTER(c_uint), POINTER(c_uint)]
    _lib.get_driver.restype = c_ubyte
    _lib.get_info.argtypes = [c_uint, c_char_p, POINTER(c_uint), c_char_p, POINTER(c_uint), c_char_p, POINTER(c_uint),
        POINTER(c_uint), c_char_p, POINTER(c_uint), c_char_p, POINTER(c_uint), POINTER(c_uint), POINTER(c_uint),
        c_char_p, POINTER(c_uint), c_char_p, POINTER(c_uint), POINTER(c_uint), POINTER(c_uint), POINTER(c_uint),
        POINTER(c_uint), POINTER(c_uint), POINTER(c_uint), c_char_p, POINTER(c_uint), POINTER(c_uint), POINTER(c_uint),
        POINTER(c_ubyte), c_char_p, POINTER(c_uint), c_char_p, POINTER(c_uint), c_char_p, POINTER(c_uint), c_char_p,
        POINTER(c_uint)]
    _lib.get_info.restype = c_ubyte
    _lib.get_status.argtypes = [c_uint, c_char_p, POINTER(c_uint), c_char_p, POINTER(c_uint), c_char_p, POINTER(c_uint),
        POINTER(c_float), POINTER(c_float), POINTER(c_float), POINTER(c_float), POINTER(c_ubyte), POINTER(c_ubyte)]
    _lib.get_status.restype = c_ubyte
    _lib.get_timeout.argtypes = [c_uint, POINTER(c_ushort)]
    _lib.get_timeout.restype = c_ubyte
    _lib.reset.argtypes = [c_uint]
    _lib.reset.restype = c_ubyte
    _lib.set_timeout.argtypes = [c_uint, c_ushort]
    _lib.set_timeout.restype = c_ubyte

    # Error List
    _error_ = {1: Error.Internal, 2: Error.Failed, 3: Error.Parameter, 4: Error.Network, 5: Error.NotFound, 6:
        Error.Timeout, 7: Error.Aborted, 8: Error.Reserved, 9: Error.Configuration}

    # Named Tuples
    _Driver = collections.namedtuple('Driver', 'manufacturer version release')
    _Info = collections.namedtuple('Info', 'model description manufacturer serial revision cpu frequency memory ' +
        'network version release calibration expiration uptime_reset uptime_total cycles eth_host ip_host ' +
        'ip_subnet_mask ip_gateway dhcp hostname domain name tag')
    _Status = collections.namedtuple('Status', 'property ok')
    _Property = collections.namedtuple('Property', 'symbol description unit typical_value min_value max_value value')


    # Configure
    @staticmethod
    def configure(serial, *args):

        # Check Parameters
        if len(args) == 0:

            # Set default Name, DHCP Status, Host IP Address, Subnet Mask, Gateway IP Address
            name = ''
            dhcp = Device.DHCP_ENABLED
            ip_host = ipaddress.ip_address('192.168.1.1')
            ip_subnet_mask = ipaddress.ip_address('255.255.255.0')
            ip_gateway = ipaddress.ip_address('0.0.0.0')

        elif (len(args) == 5):

            # Set Name, DHCP Status, Host IP Address, Subnet Mask, Gateway IP Address
            name = args[0]
            dhcp = args[1]
            ip_host = ipaddress.ip_address(args[2])
            ip_subnet_mask = ipaddress.ip_address(args[3])
            ip_gateway = ipaddress.ip_address(args[4])

        else:

            # Error
            raise Error.Parameter

        # Configure
        error = Device._lib.configure(Device._value(c_uint, serial), Device._string(name),
            Device._value(c_uint, len(name)), Device._value(c_ubyte, dhcp),
            Device._value(c_uint, int(Device._ip_address(ip_host))),
            Device._value(c_uint, int(Device._ip_address(ip_subnet_mask))),
            Device._value(c_uint, int(Device._ip_address(ip_gateway))))

        # Error Handler
        Device._error(error)


    # Get Driver
    @staticmethod
    def driver():

        # Manufacturer, Manufacturer Length, Version, Version Length, Release Date
        manufacturer = Device._string()
        manufacturer_length = Device._value(c_uint, len(manufacturer))
        version = Device._string()
        version_length = Device._value(c_uint, len(version))
        release = Device._value(c_uint)

        # Get Driver
        error = Device._lib.get_driver(manufacturer, manufacturer_length, version, version_length, release)

        # Error Handler
        Device._error(error)

        # Return Driver
        return Device._Driver(manufacturer.value.decode()[0:manufacturer_length.value],
            version.value.decode()[0:version_length.value], Device._date(release.value))


    # Find
    @staticmethod
    def find(*args):

        # Check Parameters
        if len(args) == 0:

            # Set Serial Number List
            serial = Device._array(c_uint)

        elif len(args) == 1:

            # Check Serial Number
            if isinstance(args[0], int):

                # Set Serial Number List
                serial = Device._array(c_uint, [args[0]])

            elif isinstance(args[0], set):

                # Set Serial Number List
                serial = Device._array(c_uint, args[0])

            else:

                # Error
                raise Error.Parameter

        else:

            # Error
            raise Error.Parameter

        # Address List
        address_size = Device._value(c_uint, 255)
        address = Device._array(c_uint, address_size.value)

        # Serial Number List Size
        serial_size = Device._value(c_uint, len(serial))

        # Find
        error = Device._lib.find(address, address_size, serial, serial_size)

        # Error Handler
        Device._error(error)

        # Check Parameters
        if len(args) == 0:

            # Device List
            device = set()

            # Parse Address List
            for i in range(address_size.value):

                # Insert Device into List
                device.add(Device(address[i]))

        else:

            # Check Serial Number
            if isinstance(args[0], int):

                # Set Device
                device = Device(address[0])

            else:

                # Device List
                device = {}

                # Parse Address List
                for i in range(address_size.value):

                    # Insert Device into List
                    device[serial[i]] = Device(address[i])

        # Return Device List
        return device


    # Constructor
    def __init__(self, *args):

        # Check Parameters
        if len(args) == 1:

            # Check Type
            if isinstance(args[0], Device):

                # Initialize
                self.address(args[0].address())

            else:

                # Initialize
                self.address(args[0])

        else:

            # Error
            raise Error.Parameter


    # Address
    def address(self, *args):

        # Check Parameters
        if len(args) == 0:

            # Return Address
            return ipaddress.IPv4Address(self._address.value)

        elif len(args) == 1:

            # Set Address
            self._address = self._value(c_uint, int(self._ip_address(args[0])))

        else:

            # Error
            raise Error.Parameter


    # Burn Tag
    def burn(self, tag):

        # Burn Tag
        error = self._lib.burn(self._address, self._string(tag), self._value(c_uint, len(tag)))

        # Error Handler
        self._error(error)


    # Get Info
    def info(self):

        # Model, Model Length, Description, Description Length, Manufacturer, Manufacturer Length, Serial Number,
        # Hardware Revision, Hardware Revision Length, CPU Type, CPU Type Length, CPU Core Frequency, External
        # SRAM Memory Size, Network Adapter, Network Adapter Length, Firmware Version, Firmware Version Length,
        # Firmware Release Time, Calibration Time, Expiration of Calibration Time, Uptime since Reset, Total Uptime,
        # Number of Power / Reset Cycles, Host Ethernet Address, Host IP Address, Subnet Mask, Gateway IP Address,
        # DHCP Status, Hostname, Hostname Length, Domain, Domain Length, Name, Name Length, Tag, Tag Length
        model = self._string()
        model_length = Device._value(c_uint, len(model))
        description = self._string()
        description_length = Device._value(c_uint, len(description))
        manufacturer = self._string()
        manufacturer_length = Device._value(c_uint, len(manufacturer))
        serial = self._value(c_uint)
        revision = self._string()
        revision_length = Device._value(c_uint, len(revision))
        cpu = self._string()
        cpu_length = Device._value(c_uint, len(cpu))
        frequency = self._value(c_uint)
        memory = self._value(c_uint)
        network = self._string()
        network_length = Device._value(c_uint, len(network))
        version = self._string()
        version_length = Device._value(c_uint, len(version))
        release = self._value(c_uint)
        calibration = self._value(c_uint)
        expiration = self._value(c_uint)
        uptime_reset = self._value(c_uint)
        uptime_total = self._value(c_uint)
        cycles = self._value(c_uint)
        eth_host= create_string_buffer(6)
        ip_host = self._value(c_uint)
        ip_subnet_mask = self._value(c_uint)
        ip_gateway = self._value(c_uint)
        dhcp = self._value(c_ubyte)
        hostname = self._string()
        hostname_length = Device._value(c_uint, len(hostname))
        domain = self._string()
        domain_length = Device._value(c_uint, len(domain))
        name = self._string()
        name_length = Device._value(c_uint, len(name))
        tag = self._string()
        tag_length = Device._value(c_uint, len(tag))

        # Get Info
        error = self._lib.get_info(self._address, model, model_length, description, description_length, manufacturer,
            manufacturer_length, serial, revision, revision_length, cpu, cpu_length, frequency, memory, network,
            network_length, version, version_length, release, calibration, expiration, uptime_reset, uptime_total,
            cycles, eth_host, ip_host, ip_subnet_mask, ip_gateway, dhcp, hostname, hostname_length, domain,
            domain_length, name, name_length, tag, tag_length)

        # Error Handler
        self._error(error)

        # Create and return Info
        return Device._Info(model.value.decode()[0:model_length.value],
            description.value.decode()[0:description_length.value],
            manufacturer.value.decode()[0:manufacturer_length.value], serial.value,
            revision.value.decode()[0:revision_length.value], cpu.value.decode()[0:cpu_length.value], frequency.value,
            memory.value, network.value.decode()[0:network_length.value],
            version.value.decode()[0:version_length.value], self._date(release.value), self._date(calibration.value),
            self._date(expiration.value), uptime_reset.value, uptime_total.value, cycles.value,
            self._eth_address(eth_host.raw), self._ip_address(ip_host.value), self._ip_address(ip_subnet_mask.value),
            self._ip_address(ip_gateway.value), dhcp.value, hostname.value.decode()[0:hostname_length.value],
            domain.value.decode()[0:domain_length.value], name.value.decode()[0:name_length.value],
            tag.value.decode()[0:tag_length.value])


    # Get Status
    def status(self):

        # Symbol List, Symbol List Length, Description List, Description List Length, Unit List, Unit List Length,
        # Property List Size, typical Value List, minimum Value List, maximum Value List, Value List, OK Flag
        symbol = self._string()
        symbol_length = Device._value(c_uint, len(symbol))
        description = self._string()
        description_length = Device._value(c_uint, len(description))
        unit = self._string()
        unit_length = Device._value(c_uint, len(unit))
        size = self._value(c_ubyte, 255)
        typical = self._array(c_float, size.value)
        min = self._array(c_float, size.value)
        max = self._array(c_float, size.value)
        value = self._array(c_float, size.value)
        ok = self._value(c_ubyte)

        # Get Status
        error = self._lib.get_status(self._address, symbol, symbol_length, description, description_length, unit,
            unit_length, typical, min, max, value, size, ok)

        # Error Handler
        self._error(error)

        # Split Symbol List, Description List, Unit List
        symbol = symbol.value.decode()[0:symbol_length.value].split('\t')
        description = description.value.decode()[0:description_length.value].split('\t')
        unit = unit.value.decode()[0:unit_length.value].split('\t')

        # Property List
        property = {}

        # Parse Property List
        for i in range(size.value):

            # Create Property and insert into List
            property[symbol[i]] = Device._Property(symbol[i], description[i], unit[i], typical[i], min[i], max[i],
                value[i])

        # Create and return Status
        return Device._Status(property, ok.value)


    # Reset
    def reset(self):

        # Reset
        error = self._lib.reset(self._address)

        # Error Handler
        self._error(error)


    # Timeout [ms]
    def timeout(self, *args):

        # Check Parameters
        if len(args) == 0:

            # Timeout
            timeout = self._value(c_ushort)

            # Get Timeout
            error = self._lib.get_timeout(self._address, timeout)

            # Error Handler
            self._error(error)

            # Return Timeout
            return timeout.value

        elif len(args) == 1:

            # Set Timeout
            error = self._lib.set_timeout(self._address, self._value(c_ushort, args[0]))

            # Error Handler
            self._error(error)

        else:

            # Error
            raise Error.Parameter


    # Convert List to C Array
    @staticmethod
    def _array(type, list = 0):

        # Check List
        if isinstance(list, int):

            # Create C Array
            array = (type * list)()

        else:

            # Create C Array
            array = (type * len(list))()

            # Parse List
            i = 0
            for a in list:

                # Check Conversion
                Device._value(type, a)

                # Set Element
                array[i] = a
                i += 1

        # Return C Array
        return array


    # Convert elapsed Days since 0000-01-01 to Date
    @staticmethod
    def _date(days):

        # Return Date
        return (datetime.date.fromtimestamp(min(max(days - 719528, 0), 47482) * 24 * 3600) if (days != 0) else None)


    # Convert Ethernet Address to String
    @staticmethod
    def _eth_address(eth_address):

        # Compute and return Ethernet Address
        return (':'.join('%02X' % b for b in eth_address))


    # Error Handler
    @staticmethod
    def _error(error):

        # Check Error Code
        if error != 0:

            # Raise Exception
            raise Device._error_[error]


    # Convert to IP Address
    @staticmethod
    def _ip_address(ip_address):

        # Check IP Address
        if isinstance(ip_address, str):

            try:

                # Resolve DNS and return IP Address
                return ipaddress.IPv4Address(socket.gethostbyname_ex(ip_address)[2][0]);

            except:

                # Error
                raise Error.Network

        elif isinstance(ip_address, int) and (ip_address < (1 << 32)):

            # Return IP Address
            return ipaddress.IPv4Address(ip_address)

        elif not isinstance(ip_address, ipaddress.IPv4Address):

            # Error
            raise Error.Parameter

        # Return IP Address
        return ip_address


    # Create C String
    @staticmethod
    def _string(str = ''):

        # Create and return C String
        return create_string_buffer(bytes(str, 'utf-8'), 1024)


    # Convert Value to C Value
    @staticmethod
    def _value(type, value = 0):

        try:

            # Convert to C Value
            value_ = type(value)

        except TypeError:

            # Raise Exception
            raise Error.Parameter

        # Return C Value
        return value_
