

# TMP02 - QUAD RTD MONITOR 1.2.0


# Imports
import math
from ctypes import *
from . import Device
from . import Error


# Class TMP02
class TMP02(Device):

    # Class Sensor
    class Sensor:

        # Excitation Currents [A]
        CURRENT_DEFAULT = float('nan')
        CURRENT_NONE = 0.0
        CURRENT_10UA = 0.00001
        CURRENT_1MA = 0.001

        # Nominal Resistance Limits [Ohms]
        RESISTANCE_DEFAULT = 100.0
        RESISTANCE_MIN = 1.0
        RESISTANCE_MAX = 1000.0

        # Types
        TYPE_NONE = 0
        TYPE_CUSTOM = 1
        TYPE_PT = 2
        TYPE_DT470 = 3
        TYPE_DT670 = 4
        TYPE_KTY81_1 = 5
        TYPE_KTY81_2 = 6
        TYPE_KTY82_1 = 7
        TYPE_KTY82_2 = 8
        TYPE_KTY83_1 = 9
        TYPE_KTY84_1 = 10

        # Voltage Ranges [V]
        VOLTAGE_DEFAULT = float('nan')
        VOLTAGE_0V5 = 0.5
        VOLTAGE_1V = 1.0
        VOLTAGE_2V5 = 2.5
        VOLTAGE_5V = 5.0


        # Constructor
        def __init__(self, *args):

            # Initialize
            self.voltage(self.VOLTAGE_DEFAULT)
            self.current(self.CURRENT_DEFAULT)
            self.resistance(self.RESISTANCE_DEFAULT)

            # Check Parameters
            if len(args) == 0:

                # Initialize
                self.type(self.TYPE_NONE)

            elif len(args) == 1:

                # Initialize
                self.type(args[0])

            elif len(args) == 2:

                # Initialize
                self.type(args[0])
                self.resistance(args[1])

            elif len(args) == 3:

                # Initialize
                self.type(args[0])
                self.voltage(args[1])
                self.current(args[2])

            else:

                # Error
                raise Error.Parameter


        # Excitation Current [A]
        def current(self, *args):

            # Check Parameters
            if len(args) == 0:

                # Return Current
                return self._current

            elif len(args) == 1:

                # Check Excitation Current
                if ((not math.isnan(args[0])) and (args[0] != self.CURRENT_NONE) and
                    (args[0] != self.CURRENT_10UA) and (args[0] != self.CURRENT_1MA)):

                    # Error
                    raise Error.Parameter

                # Set Excitation Current
                self._current = args[0]

            else:

                # Error
                raise Error.Parameter


        # Nominal Resistance [Ohms]
        def resistance(self, *args):

            # Check Parameters
            if len(args) == 0:

                # Return nominal Resistance
                return self._resistance

            elif len(args) == 1:

                # Check nominal Resistance
                if ((not isinstance(args[0], (int, float))) or
                    not (self.RESISTANCE_MIN <= args[0] <= self.RESISTANCE_MAX)):

                    # Error
                    raise Error.Parameter

                # Set nominal Resistance
                self._resistance = args[0]

            else:

                # Error
                raise Error.Parameter


        # Type
        def type(self, *args):

            # Check Parameters
            if len(args) == 0:

                # Return Type
                return self._type

            elif len(args) == 1:

                # Check Type
                if ((args[0] != self.TYPE_NONE) and (args[0] != self.TYPE_CUSTOM) and (args[0] != self.TYPE_PT) and
                    (args[0] != self.TYPE_DT470) and (args[0] != self.TYPE_DT670) and (args[0] != self.TYPE_KTY81_1) and
                    (args[0] != self.TYPE_KTY81_2) and (args[0] != self.TYPE_KTY82_1) and
                    (args[0] != self.TYPE_KTY82_2) and (args[0] != self.TYPE_KTY83_1) and
                    (args[0] != self.TYPE_KTY84_1)):

                    # Error
                    raise Error.Parameter

                # Set Type
                self._type = args[0]

            else:

                # Error
                raise Error.Parameter


        # Voltage Range [V]
        def voltage(self, *args):

            # Check Parameters
            if len(args) == 0:

                # Return Voltage Range
                return self._voltage

            elif len(args) == 1:

                # Check Voltage Range
                if ((not math.isnan(args[0])) and (args[0] != self.VOLTAGE_0V5) and (args[0] != self.VOLTAGE_1V) and
                    (args[0] != self.VOLTAGE_2V5) and (args[0] != self.VOLTAGE_5V)):

                    # Error
                    raise Error.Parameter

                # Set Voltage Range
                self._voltage = args[0]

            else:

                # Error
                raise Error.Parameter


    # Channels
    CHANNELS = 4

    # Sampling Frequencies [Hz]
    FREQUENCY_6 = 6
    FREQUENCY_12 = 12
    FREQUENCY_25 = 25
    FREQUENCY_50 = 50
    FREQUENCY_100 = 100
    FREQUENCY_200 = 200
    FREQUENCY_400 = 400
    FREQUENCY_800 = 800
    FREQUENCY_1500 = 1500
    FREQUENCY_3000 = 3000

    # Configure Library
    Device._lib.tmp02_measure.argtypes = [c_uint, POINTER(c_ubyte), c_ubyte, POINTER(c_float), POINTER(c_ubyte)]
    Device._lib.tmp02_measure.restype = c_ubyte
    Device._lib.tmp02_set_frequency.argtypes = [c_uint, c_ushort]
    Device._lib.tmp02_set_frequency.restype = c_ubyte
    Device._lib.tmp02_set_sensor.argtypes = [c_uint, POINTER(c_ubyte), c_ubyte, POINTER(c_ubyte), c_ubyte,
        POINTER(c_float), c_ubyte, POINTER(c_float), c_ubyte, POINTER(c_float), c_ubyte]
    Device._lib.tmp02_set_sensor.restype = c_ubyte


    # Measure [K]
    def measure(self, *args):

        # Check Parameters
        if len(args) == 0:

            # Set default Channel List
            channel = self._array(c_ubyte, range(1, self.CHANNELS + 1))

        elif len(args) == 1:

            # Check Channel
            if isinstance(args[0], int):

                # Set Channel List
                channel = self._array(c_ubyte, [args[0]])

            elif isinstance(args[0], set):

                # Set Channel List
                channel = self._array(c_ubyte, args[0])

            else:

                # Error
                raise Error.Parameter

        else:

            # Error
            raise Error.Parameter

        # Temperature List
        temperature = self._array(c_float, len(channel))

        # Measure Temperature
        error = self._lib.tmp02_measure(self._address, channel, self._value(c_ubyte, len(channel)), temperature,
            self._value(c_ubyte, len(temperature)))

        # Error Handler
        self._error(error)

        # Temperature List
        temperature_ = {}

        # Parse Channel List
        for i in range(len(channel)):

            # Insert Temperature into List
            temperature_[channel[i]] = temperature[i]

        # Return Temperature List
        return temperature_ if ((len(args) == 0) or isinstance(args[0], set)) else temperature_[channel[0]]


    # Set Sensor
    def sensor(self, *args):

        # Check Parameters
        if len(args) == 1:

            # Check Sensor
            if isinstance(args[0], self.Sensor):

                # Set default Channel List, Type List, Voltage List, Current List, Resistance List
                channel = self._array(c_ubyte, range(1, self.CHANNELS + 1))
                type = self._array(c_ubyte, [args[0].type()] * self.CHANNELS)
                voltage = self._array(c_float, [args[0].voltage()] * self.CHANNELS)
                current = self._array(c_float, [args[0].current()] * self.CHANNELS)
                resistance = self._array(c_float, [args[0].resistance()] * self.CHANNELS)

            elif isinstance(args[0], dict):

                # Parse Values
                for sensor in args[0].values():

                    # Check Sensor
                    if not isinstance(sensor, self.Sensor):

                        # Error
                        raise Error.Parameter

                # Set Channel List, Type List, Voltage List, Current List, Resistance List
                channel = self._array(c_ubyte, args[0].keys())
                type = self._array(c_ubyte, [sensor.type() for sensor in args[0].values()])
                voltage = self._array(c_float, [sensor.voltage() for sensor in args[0].values()])
                current = self._array(c_float, [sensor.current() for sensor in args[0].values()])
                resistance = self._array(c_float, [sensor.resistance() for sensor in args[0].values()])

            else:

                # Error
                raise Error.Parameter

        elif (len(args) == 2) and isinstance(args[1], self.Sensor):

            # Check Channel
            if isinstance(args[0], int):

                # Set Channel List, Type List, Voltage List, Current List, Resistance List
                channel = self._array(c_ubyte, [args[0]])
                type = self._array(c_ubyte, [args[1].type()])
                voltage = self._array(c_float, [args[1].voltage()])
                current = self._array(c_float, [args[1].current()])
                resistance = self._array(c_float, [args[1].resistance()])

            elif isinstance(args[0], set):

                # Set Channel List, Type List, Voltage List, Current List, Resistance List
                channel = self._array(c_ubyte, args[0])
                type = self._array(c_ubyte, [args[1].type()] * len(args[0]))
                voltage = self._array(c_float, [args[1].voltage()] * len(args[0]))
                current = self._array(c_float, [args[1].current()] * len(args[0]))
                resistance = self._array(c_float, [args[1].resistance()] * len(args[0]))

            else:

                # Error
                raise Error.Parameter

        else:

            # Error
            raise Error.Parameter

        # Set Sensor
        error = self._lib.tmp02_set_sensor(self._address, channel, self._value(c_ubyte, len(channel)), type,
            self._value(c_ubyte, len(type)), voltage, self._value(c_ubyte, len(voltage)), current,
            self._value(c_ubyte, len(current)), resistance, self._value(c_ubyte, len(resistance)))

        # Error Handler
        self._error(error)


    # Set Frequency [Hz]
    def frequency(self, frequency):

        # Set Frequency
        error = self._lib.tmp02_set_frequency(self._address, self._value(c_ushort, frequency))

        # Error Handler
        self._error(error)
